import logging
import typing
from threading import RLock

from scopeton.scopeTools import ScopetonException, flatten


class QualifierTree:
    def __init__(self):
        self._qualifiers = {}  # type: dict[str, typing.List[object]]
        self._qualifiersMap = {}  # type: dict[object, typing.List[str]]
        self.lock = RLock()

    def _registerObj(self, qualifier_name: str, obj):
        with self.lock:
            self._regOBjectByQualifier(qualifier_name, obj)
            self._regQualifierByObject(qualifier_name, obj)

    def _regOBjectByQualifier(self, qualifier_name, obj):
        if qualifier_name not in self._qualifiers:
            self._qualifiers[qualifier_name] = []
        if obj not in self._qualifiers[qualifier_name]:
            logging.debug("Registering: {}".format(obj))
            self._qualifiers[qualifier_name].append(obj)

    def _regQualifierByObject(self, qualifier_name, obj):
        if obj not in self._qualifiersMap:
            self._qualifiersMap[obj] = []
        if qualifier_name not in self._qualifiersMap[obj]:
            self._qualifiersMap[obj].append(qualifier_name)

    def register(self, names: typing.Union[typing.List[str], str], obj):
        if isinstance(names, str):
            names = [names]
        for name in names:
            self._registerObj(name, obj)

    def find_qualifiers(self, obj):
        return self._qualifiersMap[obj]

    def check_object_has_qualifier(self, qualifier, object):
        return qualifier in self._qualifiersMap[object]

    def get_qualifier_tree_length(self, obj):
        return len(self.find_qualifiers(obj))

    def find_suitable_qualifier(self, qualifier):
        #logging.debug("self._qualifiersMap:{}".format(self._qualifiersMap))
        #logging.debug("self._qualifiers:{}".format(self._qualifiers))
        if qualifier not in self._qualifiers:
            return qualifier
        objects = self._qualifiers[qualifier]
        if len(objects) > 1:
            raise Exception("expected 1 object for qualifier:{}, but got {}: {}".format(qualifier, len(objects), objects))
        if len(objects) == 0:
            raise Exception("No objects for qualifier:{}".format(qualifier))
        return self.find_qualifiers(objects[0])[-1]

    def contains(self, name):
        return name in self._qualifiers

    def get_all_objects(self):
        lists = flatten([self._qualifiers[q] for q in self._qualifiers])
        return list(set(lists))

    def find_by_qualifier_name(self, name):
        with self.lock:
            if name not in self._qualifiers:
                return []
            return self._qualifiers[name]

    def find_one_by_qualifier_name(self, name):
        with self.lock:
            beans = self.find_by_qualifier_name(name)
            if len(beans) == 0:
                raise ScopetonException("Error, no such qualifier: {}".format(name))
            if len(beans) > 1:
                raise ScopetonException("{} qualifier for name: {}, expected 1, objects: {}".format(len(self._qualifiers[name]), name, self._qualifiers[name]))

            return beans[0]



def flatten(lst):
    res = []
    for el in lst:
        if type(el) is list:
            res += flatten(el)
        else:
            res.append(el)
    return res

def getClassName(cls):
    return cls.__name__ if hasattr(cls, "__name__") else str(cls)

def getClassTreeQualifiers(cls):
    res = [getBean_qualifier(k) for k in getClassTree(cls)]
    res.reverse()
    return res

def rmDups(lst):
    res = []
    for k in lst:
        if k not in res:
            res.append(k)
    return res


def getClassTree(cls):
    return rmDups(_getClassTree(cls))

def _getClassTree(cls):
    res = [cls]
    if type(cls) is not object and hasattr(cls, "__bases__") and cls.__bases__:
        res += flatten([_getClassTree(parent) for parent in cls.__bases__])
    return res

def getBean_qualifier(bean):
    if isinstance(bean, str):
        return bean
    from scopeton.objects import Bean
    if isinstance(bean, Bean):
        return bean.qualifier_tree[-1]
    if not hasattr(bean, "__name__"):
        return str(bean)
    return bean.__name__

def callMethodByName(obj, name, *args, **kwargs):
    if hasattr(obj, name):
        return getattr(obj, name)(*args, **kwargs)


class ScopetonException(Exception):
    def __init__(self, *args, **kwargs):
        super(Exception, self).__init__(*args, **kwargs)


from scopeton import scopeTools
from scopeton.scopeTools import getBean_qualifier


class Bean(object):
    def __init__(self, cls, name=None, lazy = False, singleton=True, service = True, checkRegistered = True):
        self.checkRegistered = True
        self.cls = cls
        self.qualifier_tree = scopeTools.getClassTreeQualifiers(cls)
        if name:
            name = getBean_qualifier(name)
            if name not in self.qualifier_tree:
                self.qualifier_tree.append(name)
        self.lazy = lazy
        self.singleton = singleton
        self.service = service
        if not singleton and service:
            raise Exception("Error, cannot initialize service as non singleton")

    def __hash__(self):
        return hash(self.cls)

    def __eq__(self, other):
        return self.cls == other.cls


    def __repr__(self):
        return str(self)
    def __str__(self):
        name = self.__class__.__name__ if hasattr(self.__class__, "__name__") else str(self.__class__)
        return "{name}[{qualifier}]".format(self=self, name=name, qualifier=self.qualifier_tree[-1])




from scopeton import compat, glob, constants
from scopeton.scopeTools import ScopetonException


def Inject():
    def inject_decorator(fn):
        annotations_signature = compat.getMethodSignature(fn).annotations
        args_signature = compat.getMethodSignature(fn).args
        # TODO: add exception description
        def inject_wrapper(*args, **kwargs):
            scope = glob.lastScope
            nkwargs = {}
            nargs = [args[0]]
            for arg_name in args_signature:
                if arg_name == "self":
                    continue
                if arg_name not in annotations_signature:
                    raise ScopetonException("Not annotated inject argument: {}".format(arg_name))
                arg_type = annotations_signature[arg_name]
                arg_to_inject = scope.getInstance(arg_type)
                nargs.append(arg_to_inject)

            return fn(*nargs, **nkwargs)

        setattr(inject_wrapper, constants.INJECT_FLAG, 1)
        return inject_wrapper

    return inject_decorator



import logging
from threading import RLock

import typing

from scopeton import compat, glob
from scopeton.objects import Bean
from scopeton.qualifier_tree import QualifierTree
from scopeton.scopeTools import getBean_qualifier, callMethodByName, getClassTree, flatten, ScopetonException

T = typing.TypeVar("T")

class Scope(object):
    '''this is servicelocator pattern implementation'''
    def __init__(self, lock=False, initMethod="postConstruct", destroyMethod="preDestroy", parent=None):
        self._singletons = QualifierTree()
        self._beans = QualifierTree()
        self.lock = lock or RLock()       # type: RLock
        self.initMethod = initMethod
        self.destroyMethod = destroyMethod
        self.parent = parent  #type:  Scope
        self.servicesStarted = False
        self.children = []  # type: typing.List[Scope]
        self.registerInstance(self.__class__, self)
        if parent:
            parent.children.append(self)

    def getInstance(self, name: T) -> T:
        with self.lock:
            return self._getInstance(getBean_qualifier(name))

    def getInstances(self, qualifier: T) -> typing.List[T]:
        with self.lock:
            beans = self._beans.find_by_qualifier_name(getBean_qualifier(qualifier))
            beans = map(lambda x: self.getInstance(x), beans)
            return list(beans)

    def _getInstance(self, qualifier):

        suitableQualifier = self._beans.find_suitable_qualifier(qualifier)

        if self._singletons.contains(suitableQualifier):
            return self._singletons.find_one_by_qualifier_name(suitableQualifier)

        bean = self._beans.find_one_by_qualifier_name(suitableQualifier)
        glob.lastScope = self
        if compat.hasInject(bean.cls.__init__):
            instance = bean.cls()
        elif len(compat.getMethodSignature(bean.cls.__init__).args) == 2:
            instance = bean.cls(self)
        elif len(compat.getMethodSignature(bean.cls.__init__).args) > 2:
            raise ScopetonException("Invalid number of parameters for bean constructor, maybe @Inject() decorator forgotten: {}".format(compat.getMethodSignature(bean.cls.__init__).args))
        else:
            instance = bean.cls()

        if bean.singleton:
            self.registerInstance(suitableQualifier, instance)

        return instance

    def registerInstance(self, name, instance):
        qualifier = getBean_qualifier(name)
        suitableQualifier = self._beans.find_suitable_qualifier(qualifier)
        logging.debug("Suitable qualifier for {} is: {}".format(qualifier, suitableQualifier))
        self._singletons.register(suitableQualifier, instance)

    def registerBean(self, *args):
        with self.lock:
            for bean in args:
                if not isinstance(bean, Bean):
                    bean = Bean(bean)
                self._registerBean(bean)

    def remove(self):
        logging.debug("Removing scope: {}".format(self))
        self.stopServices()
        for k in self.children[:]:
            k.remove()
        if self.parent:
            self.parent.children.remove(self)


    def _registerBean(self, bean):
        """
        :type bean: Bean
        """
        for name in bean.qualifier_tree:
            logging.debug("Registering: {} as {}".format(name, bean))
            self._beans.register(name, bean)

    def runServices(self):
        if not self.servicesStarted:
            self.servicesStarted = True
            for bean in self._beans.get_all_objects():
                if bean.service:
                    callMethodByName(self.getInstance(bean), self.initMethod)

            for childScope in self.children:
                childScope.runServices()

    def stopServices(self):
        if self.servicesStarted:
            self.servicesStarted = False
            for bean in self._beans.get_all_objects():
                if bean.service:
                    callMethodByName(self.getInstance(bean), self.destroyMethod)
            for childScope in self.children:
                childScope.stopServices()

